%
%     Implementation by Piotr Hoownia <holownia@student.agh.edu.pl>
%
%     Copyright (C) 2006-10 by the HeKatE Project
%
%     PLNXT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of PLNXT.
%
%     PLNXT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     PLNXT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with PLNXT.  If not, see <http://www.gnu.org/licenses/>.
%

:- module(threads,[
			nxt_goal/2,
			nxt_goal/3,
			nxt_switch_robot/1,
			nxt_start_thread/2,
			nxt_start_thread/3,
			nxt_threads/0,
			nxt_threads/1,
			nxt_threads/2,
			nxt_threads_killall/1,
			nxt_current_robot/1,
			nxt_serial_device/1,
			nxt_actions_connection_type/1,
			nxt_sync_mode/1,
			nxt_readings/1,
			nxt_robot/10,
			trigger_create/3,
			trigger_create/4,
			trigger_create_noreturn/3,
			trigger_exists/1,
			trigger_kill/1,
			trigger_killall/0,
			trigger_killall/1,
			timer_create/3,
			timer_exists/1,
			timer_kill/1,
			timer_killall/0,
			timer_killall/1,
			wait_till/1
		]).

/** <module> Threads

Implements triggers, timers and simultaneous actions.

@author Piotr Hoownia
@license GNU General Public License
*/


% Cycle interval (time gap between two verifications of an event).
threads_cycle_interval(0.05).
% Fire interval (time gap between fireing an action, if event is still true).
threads_fire_interval(0.5)

:- dynamic nxt_running_threads/3,
	threads_trigger/2,
	threads_timer/2.



%%	nxt_goal(+Alias,+Goal).
%
%	Invokes Goal as a goal in a separated thread for a robot named Alias.

nxt_goal(Alias,Goal) :-
	nxt_goal(Alias,Goal,no_description).


%%	nxt_goal(+Alias,+Goal,+Description).
%
%	Invokes Goal as a goal in a separated thread with description for a robot named Alias.

nxt_goal(current,Goal,Description) :-
	nxt_current_robot(Alias),
	nxt_goal(Alias,Goal,Description), !.

nxt_goal(Alias,Goal,Description) :-
	thread_create(nxt_start_thread(Alias,Goal,Description),_,[detached(true),at_exit(nxt_end_thread)]).


% Thread "destructor".

nxt_end_thread :-
	thread_self(ID),
	%nxt_running_threads(_,ID,Desc),
	retractall(nxt_running_threads(_,ID,_)),
	mutex_unlock_all.
	%write('Thread ended: '),write(ID),write(' ('),write(Desc),write(').'),nl.


%%	nxt_switch_robot(Alias).
%
%	Switches current thread to work with another robot. It is very useful when controlling robots using a console.

nxt_switch_robot(Alias) :-
	thread_self(ID),
	retractall(nxt_running_threads(_,ID,console_control)),
	retractall(nxt_current_robot(_)),
	retractall(nxt_serial_device(_)),
	retractall(nxt_actions_connection_type(_)),
	retractall(nxt_sync_mode(_)),
	nxt_start_thread(Alias,true,console_control).


%%	nxt_start_thread(+Alias,+Goal).
%
%	Configures current thread to work with a robot named Alias and invokes Goal as a goal.

nxt_start_thread(Alias,Goal) :-
	nxt_start_thread(Alias,Goal,no_description).


%%	nxt_start_thread(+Alias,+Goal,+Desription).
%
%	Configures current thread to work with a robot named Alias, sets a thread description and invokes Goal as a goal.

nxt_start_thread(Alias,Goal,Description) :-
	nxt_robot_definition(Alias,Device,Conn_type,Sync_mode,Readings,WheelCircumference,AxleLenght,LeftMotor,RightMotor,PincerMotor,
		Reverse,TouchPort,SoundPort,LightPort,UltrasonicPort),
	thread_self(ID),
	assert(nxt_running_threads(Alias,ID,Description)),
	assert(nxt_current_robot(Alias)),
	assert(nxt_serial_device(Device)),
	assert(nxt_actions_connection_type(Conn_type)),
	assert(nxt_sync_mode(Sync_mode)),
	assert(nxt_readings(Readings)),
	nxt_set_robot(WheelCircumference,AxleLenght,LeftMotor,RightMotor,PincerMotor,
		Reverse,TouchPort,SoundPort,LightPort,UltrasonicPort),
	Goal.


%%	nxt_threads.
%
%	Lists all running threads.

nxt_threads :-
	writeln('------------------------------------------------------------'),
	writeln('Thread ID\tRobot\t\tDescription'),
	writeln('------------------------------------------------------------'),
	nxt_single_thread,
	writeln('------------------------------------------------------------').

nxt_single_thread :-
	nxt_running_threads(Alias,ID,Description),
	write(ID),write('\t\t'),write(Alias),write('\t\t'),write(Description),nl,
	fail.

nxt_single_thread.


%%	nxt_threads(-List).
%
%

nxt_threads(List) :-
	setof([Alias,ID,Description],nxt_running_threads(Alias,ID,Description),List).


%%	nxt_threads(+Alias,-List).
%
%

nxt_threads(current,List) :-
	nxt_current_robot(Alias),
	setof([ID,Description],nxt_running_threads(Alias,ID,Description),List).

nxt_threads(Alias,List) :-
	setof([ID,Description],nxt_running_threads(Alias,ID,Description),List).
	
	
%%	nxt_threads_killall(+Alias).
%
%	Kills all threads of robot with specified Alias (except the caller thread, the main thread, registered (asserted) triggers and registered timers).

nxt_threads_killall(Alias) :-
	not((
		nxt_running_threads(Alias, ID, _),
		not(ID = main),
		not(thread_self(ID)),
		not(threads_trigger(Alias, ID)),
		not(threads_timer(Alias, ID)),
		thread_signal(ID, thread_exit(_)),
		fail % fail in order to enforce immediate termination of all threads
		)).


%%	trigger_create(-ID,+Event,+Action).
%
%	Creates trigger.
%	Action will be fired once, when Event is true.
%	Action can be both a predicate or a list of predicates.

trigger_create(ID,Event,Action) :-
	trigger_create(ID,Event,Action,1).


%%	trigger_create(-ID,+Event,+Action,+Count).
%
%	Creates trigger.
%	Action will be fired Count times, when Event is true.
%	If Count is =inf= trigger works infinitely.
%	Action can be both a predicate or a list of predicates.

trigger_create(ID,Event,Action,Count) :-
	nxt_current_robot(Alias),
	thread_create(nxt_start_thread(Alias,trigger_start(Event,Action,Count),trigger(Event,Action)),ID,[detached(true),at_exit(nxt_end_thread)]).

trigger_start(Event,Action,Count) :-
	thread_self(ID),
	nxt_current_robot(Alias),
	assert(threads_trigger(Alias,ID)),
	trigger_cycle(ID,Event,Action,Count).

trigger_cycle(ID,Event,Action,inf) :-
	Event,
	fire(Action),
	threads_fire_interval(Interval),
	sleep(Interval),
	trigger_cycle(ID,Event,Action,inf).

trigger_cycle(ID,Event,Action,1) :-
	Event,
	fire(Action),
	retract(threads_trigger(_,ID)).

trigger_cycle(ID,Event,Action,Count) :-
	number(Count),
	Count > 1,
	Event,
	NewCount is Count - 1,
	fire(Action),
	threads_fire_interval(Interval),
	sleep(Interval),
	trigger_cycle(ID,Event,Action,NewCount).

trigger_cycle(ID,Event,Action,Count) :-
	threads_trigger(_,ID),
	threads_cycle_interval(Interval),
	sleep(Interval),
	trigger_cycle(ID,Event,Action,Count).

trigger_cycle(_,_,_,_).


%%	trigger_create_noreturn(-ID,+Event,+Action).
%
%	Creates trigger.
%	Action will be fired, when Event is true.
%	Trigger will not return to the thread which created it.
%	Action can be both a predicate or a list of predicates.

trigger_create_noreturn(ID,Event,Action) :-
	thread_self(KillID),
	nxt_current_robot(Alias),
	thread_create(nxt_start_thread(Alias,trigger_start_noreturn(KillID,Event,Action),noreturn_trigger(Event,Action)),ID,[detached(true),at_exit(nxt_end_thread)]).

trigger_start_noreturn(KillID,Event,Action) :-
	thread_self(ID),
	assert(threads_trigger(_,ID)),
	trigger_cycle_noreturn(KillID,ID,Event,Action).

trigger_cycle_noreturn(KillID,ID,Event,Action) :-
	Event,
	nxt_current_robot(Alias),
	concat(nxt_mutex_device_,Alias,Mutex),
	with_mutex(Mutex,thread_signal(KillID,thread_exit('trigger_noreturn'))),
	fire(Action),
	retract(threads_trigger(_,ID)).

trigger_cycle_noreturn(KillID,ID,Event,Action) :-
	threads_trigger(_,ID),
	threads_cycle_interval(Interval),
	sleep(Interval),
	trigger_cycle_noreturn(KillID,ID,Event,Action).

%%	trigger_exists(?ID).
%
%	Returns true if trigger exists. Fails if there is no trigger with specified ID.

trigger_exists(ID) :-
	threads_trigger(_,ID).

%%	trigger_kill(+ID).
%
%	Kills trigger with specified ID.
%	Returns true if trigger has been deleted. Fails if there is no trigger with specified ID.

trigger_kill(ID) :-
	retract(threads_trigger(_,ID)).

%%	trigger_killall.
%
%	Kills all triggers.

trigger_killall :-
	retractall(threads_trigger(_,_)).


%%	trigger_killall(+Alias).
%
%	Kills all triggers of robot with specified Alias.

trigger_killall(Alias) :-
	retractall(threads_trigger(Alias,_)).


%%	timer_create(-ID,+Time,+Action).
%
%	Creates timer.
%	Action will be fired after specified Time.
%	Action can be both a predicate or a list of predicates.

timer_create(ID,Time,Action) :-
	nxt_current_robot(Alias),
	thread_create(nxt_start_thread(Alias,timer_start(Time,Action),timer(Action)),ID,[detached(true),at_exit(nxt_end_thread)]).


timer_start(Time,Action) :-
	thread_self(ID),
	nxt_current_robot(Alias),
	assert(threads_timer(Alias,ID)),
	sleep(Time),
	% Checking if timer was killed.
	threads_timer(_,ID),
	retract(threads_timer(_,ID)),
	fire(Action).

% It has to return true.
timer_start(_,_).

%%	timer_exists(?ID).
%
%	Returns true if timer exists. Fails if there is no timer with specified ID.

timer_exists(ID) :-
	threads_timer(_,ID).

%%	timer_kill(+ID).
%
%	Kills timer with specified ID.
%	Returns true if timer has been deleted. Fails if there is no timer with specified ID.

timer_kill(ID) :-
	retract(threads_timer(_,ID)).


%%	timer_killall.
%
%	Kills all timers.

timer_killall :-
	retractall(threads_timer(_,_)).


%%	timer_killall(+Alias).
%
%	Kills all timers of robot with specified Alias.

timer_killall(Alias) :-
	retractall(threads_timer(Alias,_)).


%%	wait_till(+Event).
%
%	Waits till Event is true.

wait_till(Event) :-
	thread_self(ID),
	trigger_create(_,Event,[current_thread(ID,_),thread_send_message(ID,resume)]),
	thread_get_message(resume).


% Fires action.

fire([Action]) :- !,
	Action.

fire([Action|Tail]) :- !,
	Action,
	fire(Tail).

fire(Action) :-
	Action.
